/// <summary>Common definitions shared between attribute values and response headers</summary>
[<AutoOpen>]
module Giraffe.Htmx.Common

/// <summary>Serialize a list of key/value pairs to JSON (very rudimentary)</summary>
/// <param name="pairs">The key/value pairs to be serialized to JSON</param>
/// <returns>A string with the key/value pairs serialized to JSON</returns>
let internal toJson (pairs: (string * string) list) =
    pairs
    |> List.map (fun pair -> sprintf "\"%s\": \"%s\"" (fst pair) ((snd pair).Replace ("\"", "\\\"")))
    |> String.concat ", "
    |> sprintf "{ %s }"

/// <summary>Convert a boolean to lowercase "true" or "false"</summary>
/// <param name="boolValue">The boolean value to convert</param>
/// <returns>"true" for <c>true</c>, "false" for <c>false</c></returns>
let internal toLowerBool (boolValue: bool) =
    (string boolValue).ToLowerInvariant()


/// <summary>Valid values for the <c>hx-swap</c> attribute / <c>HX-Reswap</c> header</summary>
/// <remarks>May be combined with <c>swap</c> / <c>settle</c> / <c>scroll</c> / <c>show</c> config)</remarks>
/// <seealso href="https://htmx.org/attributes/hx-swap/">Documentation</seealso>
[<RequireQualifiedAccess>]
module HxSwap =
    
    /// <summary>The default, replace the inner HTML of the target element</summary>
    [<Literal>]
    let InnerHtml = "innerHTML"
    
    /// <summary>Replace the entire target element with the response</summary>
    [<Literal>]
    let OuterHtml = "outerHTML"
    
    /// <summary>Insert the response before the target element</summary>
    [<Literal>]
    let BeforeBegin = "beforebegin"
    
    /// <summary>Insert the response before the first child of the target element</summary>
    [<Literal>]
    let AfterBegin = "afterbegin"
    
    /// <summary>Insert the response after the last child of the target element</summary>
    [<Literal>]
    let BeforeEnd = "beforeend"
    
    /// <summary>Insert the response after the target element</summary>
    [<Literal>]
    let AfterEnd = "afterend"
    
    /// <summary>Does not append content from response (out of band items will still be processed).</summary>
    [<Literal>]
    let None = "none"
